/*
 *  wlan-control.c
 *
 *  Copyright (C) 2009 Bojan Smojver, Rexursive
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <wait.h>
#include <libgen.h>

static char lf[PATH_MAX];
static volatile sig_atomic_t sr=0,ss=0,up=0;

static void hangup(int num){
  sr=1;
}

static sig_atomic_t reconf(){
  sig_atomic_t r=sr;
  sr=0;
  return r;
}

static void suspend(int num){
  ss=1;
}

static void resume(int num){
  ss=0;
  up=1;
}

static sig_atomic_t reconn(){
  sig_atomic_t r=up;
  up=0;
  return r;
}

static void cleanup(){
  unlink(lf);
}

static int run(const char *path,...){
  pid_t pid;
  int status,nargs=0;
  const char **args,**p;
  va_list ap;

  va_start(ap,path); for(;va_arg(ap,const char*);nargs++); va_end(ap);
  args=calloc(nargs+2,sizeof(*args));
  *args=path;
  va_start(ap,path); for(p=args+1;(*p=va_arg(ap,const char*));p++); va_end(ap);

  if(!(pid=fork())){
    if(execv(path,(char *const *)args)==-1)
      exit(1);
  } else{
    waitpid(pid,&status,0);
  }

  free(args);

  return status;
}

int main(int argc,char **argv){
  int fd;
  char buf,old='1',new,*eth,*wln,lan[PATH_MAX],air[PATH_MAX],*p;
  struct sigaction act;
  sig_atomic_t up;

  if(argc<3)
    return 1;

  snprintf(lf,sizeof(lf),"/var/lock/%s",basename(p=strdup(*argv)));
  free(p);

  if((fd=open(lf,O_CREAT|O_EXCL|O_WRONLY,0))==-1)
    return 1;
  close(fd);

  atexit(cleanup);

  memset(&act,0,sizeof(act));
  act.sa_handler=hangup;
  sigaction(SIGHUP,&act,NULL);

  act.sa_handler=suspend;
  sigaction(SIGUSR1,&act,NULL);
  act.sa_handler=resume;
  sigaction(SIGUSR2,&act,NULL);

  act.sa_handler=exit;
  sigaction(SIGINT,&act,NULL);
  sigaction(SIGQUIT,&act,NULL);
  sigaction(SIGTERM,&act,NULL);

  eth=*(argv+1);
  wln=*(argv+2);

  snprintf(lan,sizeof(lan),"/sys/class/net/%s/carrier",eth);
  snprintf(air,sizeof(air),"/sys/class/net/%s/carrier",wln);

  while(1){
    if(ss) goto sleep;

    if((up=reconn()))
      sleep(30);

    if((fd=open(lan,O_RDONLY))==-1){
      run("/sbin/ip","link","set","dev",eth,"up",NULL);
      if((fd=open(lan,O_RDONLY))==-1)
        goto sleep;
    }
    if(read(fd,&new,sizeof(new))!=sizeof(new)){
      run("/sbin/ip","link","set","dev",eth,"up",NULL);
      if(read(fd,&new,sizeof(new))!=sizeof(new)){
        close(fd);
        goto sleep;
      }
    }
    close(fd);

    if(new!=old || reconf() || (up && new=='0')){
      switch(new){
      case '1':
        run("/sbin/ifdown",wln,NULL);
        run("/sbin/service","wpa_supplicant","stop",NULL);
        run("/sbin/ifdown",eth,NULL);
        run("/sbin/ifup",eth,NULL);
        break;
      case '0':
        run("/sbin/ifdown",eth,NULL);
        run("/sbin/ifdown",wln,NULL);
        run("/sbin/service","wpa_supplicant","restart",NULL);
        sleep(10);
        run("/sbin/ifup",wln,NULL);

        if((fd=open(air,O_RDONLY))==-1){
          run("/sbin/ifdown",wln,NULL);
          run("/sbin/service","wpa_supplicant","stop",NULL);
          goto out;
        }
        if(read(fd,&buf,sizeof(buf))!=sizeof(buf) || buf=='0'){
          close(fd);
          run("/sbin/ifdown",wln,NULL);
          run("/sbin/service","wpa_supplicant","stop",NULL);
          goto out;
        }
        close(fd);

        break;
      default:
        break;
      }
out:
      old=new;
    }

sleep:
    sleep(10);
  }

  return 0;
}
